import sys, random, math, pygame
from time import time
from .Sprite import Sprite
from .constants import *
from .Globals import Globals
from .utils import at, get, is_wall, get_animation
from .Inventory import *
from functools import partial

# actor class for four directional movement
class Actor(Sprite):
    MAX_WEALTH = 500000
    def __init__(self, actions, rect, tag=None, abortable=False, name=None):
        # - scale images
        self.actions = {}
        self.masks = {}
        for action in actions:
            self.actions[action] = []
            self.masks[action] = []
            for img in actions[action]:
                img = img.convert_alpha()
                img = pygame.transform.scale(img, rect.size)
                self.actions[action].append(img)
                self.masks[action].append(pygame.mask.from_surface(img))


        self.index = 0
        self.action_iterations = 0
        self.action = IDLE
        self.action_loop = FOREVER

        self._stop = False
        self.frame_count = 0
        self.frame_rate = 1
        self.prev_vector = None
        self.direction = LEFT

        self._defend = None
        self._health = 100.0
        self._wealth = 500.0
        self._energy = 100.0
        self._inventory = None
        self.attributes = {}

        if tag is not None and tag in Globals.cache:
            self._inventory = Globals.cache[tag]
        else:
            self._inventory = Inventory()
            Globals.cache[tag] = self._inventory
        self._inventory.actor = self

        surface = actions[self.action][self.index]
        Sprite.__init__(self, surface, rect, tag, abortable, name)

    @property
    def health(self):
        return self._health

    @health.setter
    def health(self, value):
        if value >= 0 and value < 100:
            self._health = value
        if self.health == 0:
            self.act(DIE, loop=1)

    @property
    def energy(self):
        return self._energy

    @energy.setter
    def energy(self, value):
        self._energy += value
        if self._energy < 0: self._energy = 0
        if self._energy > 100: self._energy = 100

    @property
    def wealth(self):
        return self._wealth

    @wealth.setter
    def wealth(self, value):
        self._wealth += value
        if self._wealth > Actor.MAX_WEALTH: self._wealth = Actor.MAX_WEALTH

    @property
    def defend(self):
        return self._defend

    @defend.setter
    def defend(self, callback):
        self._defend = callback

    def stop(self):
        Sprite.stop(self)
        self.act(IDLE, FOREVER)
        self._stop = True

    def move(self, vector, **kwargs):
        animation = kwargs.get('animation', WALK + '_' + self.direction)
        action = kwargs.get('action', WALK)
        #speed = kwargs.get('speed', None)

        if self.health == 0:
            return

        direction = None
        if action == WALK:
            if vector[0] > 0:
                direction = RIGHT
            elif vector[0] < 0:
                direction = LEFT
            elif vector[1] > 0:
                direction = FRONT
            elif vector[1] < 0:
                direction = BACK

        if direction is not None:
            if direction != self.direction:
                self.index = 0
                if action == WALK:
                   animation = WALK + '_' + direction
            self.direction = direction

        self.act(animation, FOREVER)
        Sprite.move(self, vector, **kwargs)

    def move_to(self, *points, **kwargs):
        callback = kwargs.get('callback', None)
        pabort = kwargs.get('pabort', -1)
        action = kwargs.get('action', WALK)
        animation = kwargs.get('animation', WALK + '_' + self.direction)
        speed = kwargs.get('speed', None)
        if self._stop:
           self._stop = False
        elif len(points) == 0:
           self.act(IDLE, FOREVER)
        elif random.uniform(0, 1) <= pabort:
           self.act(IDLE, FOREVER)
        else:
           if len(points) > 1:
              head, *tail = points
              if is_wall(head):
                 self.act(IDLE, FOREVER)
              else:
                 self.move((head[0]-self.x, head[1]-self.y), callback=partial(self.move_to, *tail, **kwargs), action=action, animation=animation, speed=speed)
           else:
              head = points[0]
              if is_wall(head):
                 self.act(IDLE, FOREVER)
              else:
                 if callback is not None:
                    self.move((head[0]-self.x, head[1]-self.y), callback=callback, action=action, animation=animation, speed=speed)
                 else:
                    self.move((head[0]-self.x, head[1]-self.y), callback=partial(self.act, IDLE+'_'+self.direction, FOREVER), action=action, animation=animation, speed=speed)
        return self

    def jump(self, height = 4, arc = [1.875, 3.875, 3.875, 1.875]):

       ani = get_animation(self)
       if (ani is not None and (ani.action == JUMP or ani.action == GRAVITY)):
          return

       if ani is not None and ani.action == WALK:
          inc = 1
          if self.direction == LEFT:
             inc = -1
          pts = []

          for x in range(len(arc)):
              pts.append((self.x+ inc + (x*inc), self.y - arc[x]))
          pts.append((self.x + inc + (inc * len(arc)), self.y))
          Globals.instance.animations.remove(ani)
          self.move_to(*pts, animation=WALK + '_' + self.direction, action=JUMP, speed=30, callback=ani.callback)

       else:
          self.move_to((self.x, self.y-height),(self.x, self.y), speed=30, animation=WALK + '_' + self.direction, action=JUMP)

    def _complete_move(self, callback = None):
        if self.health == 0:
            return

        self.act(IDLE + '_' + self.direction, FOREVER)
        Sprite._complete_move(self, callback)

    def _update(self, delta):

        # check to see if we need to activate self defense
        if self._defend is not None:
            self._defend(self)

        img = self.actions[self.action][self.index]
        self.surface = img
        self.origin_surface = img
        self.mask = self.masks[self.action][self.index]
        Sprite._update(self, delta)
        if self.action_loop == FOREVER or self.action_iterations < self.action_loop:
            self.frame_count = self.frame_count + 1
            if self.frame_count >= self.frame_rate:
                self.index = self.index + 1
                self.frame_count = 0
            if self.index >= len(self.actions[self.action]):
                self.action_iterations = self.action_iterations + 1
                if self.action_loop == FOREVER or self.action_iterations < self.action_loop:
                    self.index = 0
                else:
                    self.index = self.index - 1
        elif self.health > 0:
            self.index = 0
            self.action = IDLE + '_' + self.direction
            self.action_loop = FOREVER

    def get_actions(self):
        """ return all actions that this actor can perform """
        return list(self.actions.keys())


    def take(self, obj):
        """ take this object and add to inventory """
        self._inventory.add(obj)
        obj.actor = self
        return self

    # TODO: this needs to be merged with act
    def actit(self, action, loop=FOREVER):
        if not action in self.actions:
            print('Unsupported action ' + str(action) + '. Going to IDLE')
            self.act(IDLE, loop)
            return self
            #print('Unsupported action ' + str(action) + '. Valid options are:')
            #for action in self.actions:
            #    print(action.upper())
            #sys.exit(0)
        self.index = 0
        self.action = action
        self.action_loop = loop
        self.action_iterations = 0
        return self

    def act(self, action, loop=FOREVER):
        if action == GRAVITY or action == None:
           action = IDLE + '_' + self.direction
        if self.health > 0 or action.startswith(DIE):
            if action in self.actions:
                self.actit(action, loop)
            else:
                # infer the action based on direction
                if str(action + '_' + self.direction) in self.actions:
                    self.actit(str(action + '_' + self.direction), loop)
                else:
                    self.actit(action, loop)
        return self

    def kill(self, delay=1.25):
        """ used to kill this actor """
        if self.health > 0:
            self.stop()
            self.health = 0
            self.destruct(delay)

    def rate(self, frame_rate):
        """ the rate to swap animation frames, default is 1 per update call """
        if frame_rate < 0:
            frame_rate = 1
        if frame_rate > 60:
            frame_rate = 60
        self.frame_rate = frame_rate
        return self

    def facing(self, distance=100):
        """ returns a position (off the screen) where this actor is facing """
        if self.direction == BACK:
            return self.x, self.y - distance
        elif self.direction == FRONT:
            return self.x, self.y + distance
        elif self.direction == LEFT:
            return self.x - distance, self.y
        elif self.direction == RIGHT:
            return distance + self.x, self.y

    def next(self):
        return next(self.direction)

    def next(self, direction):
        """ the next position (in the current direction or relative board position) """
        if direction == BACK:
            return self.x, self.y - 1
        elif direction == FRONT:
            return self.x, self.y + 1
        elif direction == LEFT:
            return self.x - 1, self.y
        elif direction == RIGHT:
            return self.x + 1, self.y

    def _check_next_object(self, pos):
        lst = at(pos)
        if lst is not None:
            if isinstance(lst, list):
                for x in lst:
                    if x != self and isinstance(x, Actor):
                        if x.health > 0:
                            return x
                    elif x != self and isinstance(x, Sprite):
                        return x
            elif lst != self and isinstance(lst, Actor):
                if lst.health > 0:
                    return lst
            elif lst != self and isinstance(lst, Sprite):
                return lst

    def next_object(self, direction=None, distance=50):
        """ returns the next thing along the path where this actor is facing,
            up to a certain distance. if the next thing is an Actor, it must be alive.
        """
        if direction is None: direction = self.direction
        if direction == BACK:
            for y in range(self.y, self.y - distance, -1):
                obj = self._check_next_object((self.x, y))
                if obj is not None:
                    return obj
        elif direction == FRONT:
            for y in range(self.y, self.y + distance, 1):
                obj = self._check_next_object((self.x, y))
                if obj is not None:
                    return obj
        elif direction == LEFT:
            for x in range(self.x, self.x - distance, -1):
                obj = self._check_next_object((x, self.y))
                if obj is not None:
                    return obj
        elif direction == RIGHT:
            for x in range(self.x+1, self.x + distance, 1):
                obj = self._check_next_object((x, self.y))
                if obj is not None:
                    return obj

    def dump_state(self):
        """ extract wealth, health, energy, inventory """
        d = {}
        d['wealth'] = self._wealth
        d['health'] = self._health
        d['energy'] = self._energy
        d['inventory'] = self._inventory.dump_state()
        return d

    def load_state(self, state):
        """ load saved state values """
        if 'wealth' in state:
            self._wealth = state['wealth']
        if 'health' in state:
            self._health = state['health']
        if 'energy' in state:
            self._energy = state['energy']
        if 'inventory' in state:
            self._inventory.load_state(state['inventory'])



# something an Actor can use
from functools import partial
from .utils import register_keydown as keydown, at, get, has_tag, sprites
from .constants import *

class Thing:
    def __init__(self, call=None):
        self.name = None
        self.image = None
        self.lethality = 100
        self.energy = 1
        self.quantity = 1
        self.actor = None
        self.cost = 1000

        if call is not None:
            keydown(call, self.use)

    def use(self):
        raise NotImplementedError('base class cannot be called directly')

    def __str__(self):
        return '{} {}'.format(self.name, self.quantity)



import random, math, os, sys
from types import ModuleType
from .Globals import Globals
from .Animation import Animation
from .constants import *
from astar import AStar
from functools import partial
from math import ceil
from bresenham import bresenham
import copy

def load_module(path, api):
    src = open(path).read()
    code = compile(src, os.path.basename(path), 'exec', dont_inherit = True)

    name, _ = os.path.splitext(os.path.basename(path))
    mod = ModuleType(name)
    if api is not None:
       mod.__dict__.update(api.__dict__)
    sys.modules[name] = mod

    return code, mod

def import_plugin(plugin_file):
    """ allows other 'plugin' code to be imported into a game. this works a little like load module """
    from . import api
    code, mod = load_module(plugin_file, api)
    exec(code, mod.__dict__)
    return mod

def register_cell(pos, s):
    """ helper function that builds the index of all sprites in a given cell """
    area = to_area(s.x, s.y, s.width, s.height)
    for p in area:
        lst = []
        if p in Globals.instance.cells:
            lst = Globals.instance.cells[p]
        else:
            Globals.instance.cells[p] = lst
        lst.append(s)

def register_keydown(key, callback):
    # single key callbacks
    Globals.instance.keys_registered['keydown'][key] = set([callback])
    #if key in Globals.instance.keys_registered['keydown']:
    #    Globals.instance.keys_registered['keydown'][key].add(callback)
    #else:
    #    Globals.instance.keys_registered['keydown'][key] = set([callback])

def register_keyup(key, callback):
    # single key callbacks
    Globals.instance.keys_registered['keyup'][key] = set([callback])

    #if key in Globals.instance.keys_registered['keyup']:
    #    Globals.instance.keys_registered['keyup'][key].add(callback)
    #else:
    #    Globals.instance.keys_registered['keyup'][key] = set([callback])

def has_animation(obj, action=GRAVITY):
    for o in Globals.instance.animations:
        if o.obj == obj and o.action == action:
            return True
    else:
        return False



def get_animation(obj):
    for o in Globals.instance.animations:
        if o.obj == obj:
            return o


def animate(obj, duration = 1, callback = None, abortable=False, action=None, **kwargs):
    Globals.instance.animations.append(Animation(obj, duration, callback, abortable, action, **kwargs))

def at(pos):
    if pos in Globals.instance.cells:
        lst = Globals.instance.cells[pos]
        if len(lst) == 0:
            return None
        elif len(lst) == 1:
            return lst[0]
        else:
            return lst

def to_grid(screen_pos):
    """ converts screen coordinates to grid coordinates """
    if isinstance(screen_pos, list):
        return tuple((int((x / (Globals.instance.GRID_SIZE * 1.0))) for x in screen_pos))
    else:
        return int(screen_pos / (Globals.instance.GRID_SIZE * 1.0))

def to_area(x, y, w, h, bottom_only=False):
    """ takes a upper_left point, width and height and returns full covering area (as list) """
    cover = []
    if bottom_only:
        for i in range(int(ceil(w))):
            cover.append((int(x)+i, int(y+ceil(h)-1)))
    else:
        for i in range(int(ceil(w))):
            for j in range(int(ceil(h))):
                cover.append((int(x)+i, int(y)+j))

    return cover

def is_wall(pos, cells=None):
    """ returns true if a wall is at the desired location, false otherwise """
    atpos = None
    if cells is None:
        atpos = at(pos)
    else:
        atpos = cells
    if atpos is not None:
        if isinstance(atpos, list):
            for x in atpos:
                if x.tag == 'wall':
                    return True
        elif atpos.tag == 'wall':
            return True

    return False

def has_obstacle(lst):
    """ check to see if this list has a sprite that is an obstacle """
    for x in lst:
        if x.tag is not None and x.tag == OBSTACLE:
            return True
    else:
        return False

def max_distance(posA, posB, width, height, skip_up=True):
    """ see if it is possible to move between two points (returns the furthest distance) """
    x, y = posA
    x_dest, y_dest = posB

    cross = list(bresenham(x, y, x_dest, y_dest))[1:]
    x_dest = x_prev = x
    y_dest = y_prev = y
    #print('{} ==> {}'.format(posA, posB))
    #print('cross {}'.format(cross))
    #print('cells {}'.format(Globals.instance.cells))

    for p in cross:
        # see if any the next points will hit an obstacle (coming down)
        next_area = to_area(p[0], p[1], width, height, bottom_only=True)
        #print(' next area {}'.format(next_area))
        clear = True
        if p[1] > y_prev or skip_up is False:
            for pn in next_area:
                if pn in Globals.instance.cells and is_wall(pn, cells=Globals.instance.cells[pn]):
                    #print(' there is something at {}'.format(pn))
                    clear = False
                    break
        if clear:
            x_dest = int(p[0])
            y_dest = int(p[1])
        else:
            break
        x_prev = p[0]
        y_prev = p[1]
    #print('jumping to {}'.format((x_dest,y_dest)))
    return x_dest, y_dest

def get(name):
    if name in Globals.instance.tags:
        return copy.copy(Globals.instance.tags[name])
    else:
        return []

def has_tag(lst, tag):
    """ checks if a list contains a sprite with a given tag """
    if not isinstance(lst, list):
        lst = [lst]
    for l in lst:
        if l.tag == tag:
            return True
    else:
        return False

def rand_pos(x_padding = 0, y_padding = 0, empty=False):
    grid_width = (Globals.instance.WIDTH / Globals.instance.GRID_SIZE) - math.ceil(x_padding)
    grid_height = (Globals.instance.HEIGHT / Globals.instance.GRID_SIZE) - math.ceil(y_padding)
    x = 0
    y = 0
    while True:
        x = random.randrange(0, grid_width)
        y = random.randrange(0, grid_height)

        if len(Globals.instance.sprites) >= grid_width * grid_height:
            break

        if at((x, y)) is None:
            break

    return x, y

def rand_maze(callback):
    from .Maze import Maze
    w = int(Globals.instance.WIDTH/Globals.instance.GRID_SIZE)
    h = int(Globals.instance.HEIGHT/Globals.instance.GRID_SIZE)
    nw = int((w-1)/2)
    nh = int((h-1)/2)
    maze = Maze.generate(nw, nh)
    rows = maze._to_str_matrix()
    for rnum in range(len(rows)):
        row = rows[rnum]
        for cnum in range(len(row)):
            if row[cnum] == 'O':
                s = callback(pos=(cnum,rnum), tag=OBSTACLE)
                register_cell((cnum, rnum), s)

def rand_color():
    r = random.randrange(0, 255)
    g = random.randrange(0, 255)
    b = random.randrange(0, 255)
    #if (r, g, b) == globs.BACKGROUND:
    #    r, g, b = rand_color()
    return r, g, b

def rand_arc():
    p1 = rand_pos(1,4)
    p2 = rand_pos(1,4)
    mid_x = (p1[0] + p2[0]) / 2
    return (p1[0], (Globals.instance.HEIGHT/Globals.instance.GRID_SIZE)+1), (int(mid_x), 1), (p2[0], (Globals.instance.HEIGHT/Globals.instance.GRID_SIZE)+1)

def roundup(num, step):
    return int(math.ceil(num / float(step))) * step

def randrange_float(start, stop, step):
    return random.randint(0, int((stop - start) / step)) * step + start

def sign(num):
    return (1, -1)[num < 0]

def distance(p1, p2):
    return math.sqrt(sum([(a - b) ** 2 for a, b in zip(p1, p2)]))

def visible(p1):
    if p1[0] >= 0 and p1[1] >= 0 and p1[0] < (Globals.instance.WIDTH/Globals.instance.GRID_SIZE) and p1[1] < (Globals.instance.HEIGHT/Globals.instance.GRID_SIZE):
        return True
    else:
        return False

def score_pos(pos = UPPER_LEFT):
    """ return the grid position of the score sprite """
    return {
        UPPER_LEFT : (0.5, 0.5),
        UPPER_RIGHT: ((Globals.instance.WIDTH/Globals.instance.GRID_SIZE) - 0.5, 0.5),
        LOWER_LEFT:  (0.5, (Globals.instance.HEIGHT/Globals.instance.GRID_SIZE) - 1),
        LOWER_RIGHT: ((Globals.instance.WIDTH/Globals.instance.GRID_SIZE) - 0.5, (Globals.instance.HEIGHT/Globals.instance.GRID_SIZE) - 1)
    }.get(pos, UPPER_LEFT)

def sprites():
    """ return a list of all loaded sprites """
    return Globals.instance.sprites

def graze(sprite) :
    """ a sprite.wander() operation. randomly move around """
    x, y = sprite.pos
    choices    = [(x,y), (x, y-1), (x, y+1), (x+1, y), (x-1, y)]
    random.shuffle(choices)
    obstacles  = [at(p) for p in choices]
    visibility = [visible(p) for p in choices]

    for i in range(len(choices)):
        if obstacles[i] is None and visibility[i]:
            if choices[i] != (x, y):
                sprite.move((choices[i][0] - x, choices[i][1] - y))
                break

class __GridSolver__(AStar):
    def __init__(self):
        self.xx = 1

    def heuristic_cost_estimate(self, n1, n2):
        (x1, y1) = n1
        (x2, y2) = n2
        return math.hypot(x2 - x1, y2 - y1)

    def distance_between(self, n1, n2):
        return 1

    def neighbors(self, node):
        x, y = node
        choices = [(x, y - 1), (x, y + 1), (x - 1, y), (x + 1, y)]
        valid = [visible(p) is True and is_wall(p) is False for p in choices]
        ret = []
        for i in range(len(valid)):
            if valid[i] is True :
               ret.append(choices[i])
        return ret

def track_astar(sprite, find_tags, pabort=-1):
    enemies = []
    for t in find_tags:
       enemies.extend(get(t))

    if len(enemies) == 0:
       sprite.act(IDLE, FOREVER)
       return

    enemy = random.choice(enemies)
    x, y = sprite.pos
    path = __GridSolver__().astar(sprite.pos,enemy.pos)
    if path is not None:
        lst = list(path)
        if len(lst) > 0:
           sprite.move_to(*lst, pabort=pabort)
        else:
           sprite.act(IDLE, FOREVER)
    else:
        sprite.act(IDLE, FOREVER)

def track(sprite, find_tags, pbad = 0.1) :
    """
        a sprite.wander() operation. attempt to a path that moves sprite closer to player_sprite.

        :param sprite: the sprite to automate movements.

        :param find_tags: the tags to track

        :param pbad: the probability to make a bad move. some number of bad moves are needed to

    """

    enemies = []
    for t in find_tags:
       enemies.extend(get(t))

    distances = [distance(e.pos, sprite.pos) for e in enemies]

    enemy = enemies[distances.index(min(distances))]
    x, y = sprite.pos

    choices    = [(x, y), (x, y-1), (x, y+1), (x+1, y), (x-1, y)]
    distances  = [distance(p, enemy.pos) for p in choices]
    visibility = [visible(p) for p in choices]

    best = None
    min_dist = 999999
    for i in range(len(choices)):
        if is_wall(choices[i]) or not visibility[i]:
            continue

        #every now and then make a random "bad" move
        rnd = random.uniform(0, 1)
        if rnd <= pbad:
            best = choices[i]
            break
        elif distances[i] < min_dist:
            best = choices[i]
            min_dist = distances[i]
    if best is not None and best != (x,y):
        sprite.move((best[0] - x, best[1] - y))

def player_physics(action, sprite, vector):
    """ simple physical model that keep a player from walking into obstacles """
    area = []
    if ceil(sprite.width) > 1 or ceil(sprite.height) > 1:
        area = to_area(sprite.x, sprite.y, sprite.width, sprite.height)
    else:
        area.append(sprite.pos)

    # shift each point by the vector
    area = list((x + vector[0], y + vector[1]) for x,y in area)

    for pos in area:
        obj = at(pos)
        if obj and isinstance(obj, list):
            for x in obj:
                if x.tag == OBSTACLE:
                    return False
        elif obj and obj.tag == OBSTACLE:
            return False
        elif not visible(pos):
            return False
    return True

def fill(obj, prob = 1, collide_obj = None, collide_callback = None) :
    """
        fills some amount white space with an object. object can be set to collide with something (collide_obj) in which case a callback would be invoked (collide_callback).

        :param prob: the probability of filling a given square with an object.

        :param obj: a callback (or partial) to a sprite to create in whitespace.

        :param collide_obj: an object or (list of objects) to check for collisions

        :param collide_callback: a callback function to invoke when collide_obj collides with obj.

    """
    for x in range(int(Globals.instance.WIDTH/Globals.instance.GRID_SIZE)):
        for y in range(int(Globals.instance.HEIGHT/Globals.instance.GRID_SIZE)):
             if random.uniform(0, 1) > prob:
                continue
             if at((x,y)) is None:
                o = obj(pos=(x,y))
                if collide_obj and collide_callback:
                    if isinstance(collide_obj, (list, tuple)):
                        for obj in collide_obj:
                            o.collides(obj, collide_callback)
                    else:
                        o.collides(collide_obj, collide_callback)

def load_state(actor, sfile):
    """
        save the state of an actor in an encoded file

        :param actor: actor class reference to load state into

        :param sfile: encoded file containing state information
    """
    import os.path, json, base64
    if os.path.isfile(sfile):
        state_file = open(sfile, 'rb')
        state = json.loads(base64.b64decode(state_file.read()).decode("ascii"))
        actor.load_state(state)
        return True
    return False

def save_state(actor, sfile):
    """
        loads previously saved actor state from encoded file

        :param actor: actor class to dump state from

        :param sfile: encoded file used to write state information
    """
    import json, base64
    state_file = open(sfile, 'wb')
    state_file.write(base64.b64encode(json.dumps(actor.dump_state()).encode()))

def vadd(x,y):
    return [x[0]+y[0],x[1]+y[1]]

def vsub(x,y):
    return [x[0]-y[0],x[1]-y[1]]

def vdot(x,y):
    return x[0]*y[0]+x[1]*y[1]



from .Globals import Globals
from .constants import *
import math
class Animation:
    def __init__(self, obj, duration, callback, abortable=False, action=None, **kwargs):
        self.obj = obj
        self.start = {}
        self.attributes = kwargs
        for attribute in self.attributes:
            self.start[attribute] = getattr(obj, attribute)
        self.time = float(0)
        self.duration = float(duration)

        # prevent div by 0
        if self.duration == 0.0:
            self.duration = 0.00001

        self.callback = callback
        self.finished = False
        self.abortable = abortable
        self.action = action
        self.acceleration = 0.000

    def update(self, delta):
        self.time += (delta / 1000)
        if self.action == GRAVITY:
           #print('adding gravity {} from {} {}'.format(self.acceleration * self.obj.mass, self.acceleration, self.obj.mass))
           self.time += self.acceleration * self.obj._mass
        n = self.time / self.duration

        if n >= 1:
            self.finished = True
            n = 1

        for attribute in self.attributes:
            step = self.attributes[attribute] - self.start[attribute]
            #print('{} : {} --> {} [{}]'.format(attribute, getattr(self.obj, attribute), n * step + self.start[attribute], n * step + self.start[attribute]-getattr(self.obj, attribute)))
            setattr(self.obj, attribute, n * step + self.start[attribute])

    def abort(self):
        if self.abortable:
            self.finished = True
            if self.progress() > 50:
                for attribute in self.attributes:
                    setattr(self.obj, attribute, self.attributes[attribute])

    def progress(self):
        min_progress = 100.0

        for attribute in self.attributes:
            if self.attributes[attribute] == 0:
                continue
            progress = math.fabs(100.0 - ((self.attributes[attribute] - getattr(self.obj, attribute))/(self.attributes[attribute])))
            if progress < min_progress:
                min_progress = progress
        return min_progress

    def finish(self):
        if not self.obj in Globals.instance.sprites:
            return

        if self.callback:
            self.callback()



import sys, os, shutil, json
import urllib.request, traceback
from io import BytesIO
from github import Github
from zipfile import ZipFile
from types import ModuleType
from . import predigame
from .utils import load_module
from predigame.constants import *

from pkg_resources import get_distribution
__version__ = get_distribution('predigame').version

def bootstrap():
    if len(sys.argv) == 1:
        print('Predigame Instructional Platform\n')
        print('Running a Game:')
        print('   pred some_file.py\n')
        print('Create a New Game:')
        print('   pred new some_game\n')
        print('List Available Game Downloads:')
        print('   pred list\n')
        print('Download a Game:')
        print('   pred pull some_game')
        print('   pred fetch some_game\n')

        sys.exit()
    if sys.argv[1] == 'new':
        new_game()
    elif sys.argv[1] == 'list':
        get_games()
    elif sys.argv[1] == 'pull':
        pull_game()
    elif sys.argv[1] == 'fetch':
        fetch_game()
    else:
        main()

def err():
    print('Error: Invalid Python file provided')
    sys.exit()

def main():
    try:
        run_mod = sys.argv[1:][0]
    except:
        err()

    path = os.path.join(os.getcwd(), run_mod)
    from . import api
    code, mod = load_module(path, api)

    dummy_mod = ModuleType('dummy')

    try:
        exec(code, dummy_mod.__dict__)
    except:
        pass
    finally:
        WIDTH = getattr(dummy_mod, 'WIDTH', 16)
        HEIGHT = getattr(dummy_mod, 'HEIGHT', 16)
        TITLE = getattr(dummy_mod, 'TITLE', 'Predigame')
        SIZE = getattr(dummy_mod, 'SIZE', 50)
        BACKGROUND = getattr(dummy_mod, 'BACKGROUND', (220, 220, 220))
        FULLSCREEN = getattr(dummy_mod, 'FULLSCREEN', False)
        COLLISIONS = getattr(dummy_mod, 'PIXEL_COLLISIONS', True)

    predigame.init(path, WIDTH * SIZE, HEIGHT * SIZE, TITLE, grid = SIZE, bg = BACKGROUND, fullscreen = FULLSCREEN, collisions = COLLISIONS)

    exec(code, mod.__dict__)

    while True:
        predigame.main_loop()

def pull_game():
    if len(sys.argv) != 3:
        print('Usage: pred pull <game>')
        sys.exit()
    game = sys.argv[2]

    if os.path.exists(game):
        prompt = input('{} already exists. Overwrite? (Y or N): '.format(game))
        if prompt.upper() == 'Y':
            shutil.rmtree(game)
        else:
            sys.exit()
    try:
        g = Github()
        repo = g.get_organization('predigame').get_repo(game)
        tags = repo.get_tags()
        tag_url = None
        for tag in tags:
            tag_name = tag.name
            tag_url = tag.zipball_url
            if tag_name == __version__:
                tag_url

        if tag_url is not None:
            print('Fetching game {} from {}'.format(game, tag_url))
            with urllib.request.urlopen(tag_url) as response:
                data = response.read()
                with ZipFile(BytesIO(data)) as dnld:
                    prefix = dnld.namelist()[0].split('/')[0]
                    dnld.extractall()
                    os.rename(prefix, game)
            print('Download Complete!')
    except:
        print('Unable to pull game {}. Does it exist?'.format(game))
        traceback.print_exc()

def fetch_game():
    """

    Similar to pull game but doesn't use the Github API and avoids stumbling into SSL verification issues.

    """
    if len(sys.argv) != 3:
        print('Usage: pred fetch <game>')
        sys.exit()
    game = sys.argv[2]

    if os.path.exists(game):
        prompt = input('{} already exists. Overwrite? (Y or N): '.format(game))
        if prompt.upper() == 'Y':
            shutil.rmtree(game)
        else:
            sys.exit()

    url = 'http://predigame.io/games/' + game + '.zip'
    try:
        with urllib.request.urlopen(url) as response:
            data = response.read()
            with ZipFile(BytesIO(data)) as dnld:
                prefix = dnld.namelist()[0].split('/')[0]
                dnld.extractall()
                os.rename(prefix, game)
            print('Download Complete!')
    except:
        print('Unable to fetch game {} from {}. Does it exist?'.format(game, url))
        traceback.print_exc()


def get_games():
    g = Github()
    repos = g.get_organization('predigame').get_repos()
    for repo in repos:
        name = repo.full_name.replace('predigame/', '')
        if name == 'predigame':
            continue
        desc = repo.description
        tags = repo.get_tags()
        version_match = False
        for tag in tags:
            tag_name = tag.name
            tag_url = tag.zipball_url
            if tag_name == __version__:
                version_match = True
            #print('  {} \t {}'.format(tag_name, tag_url))
        if version_match:
            print('{0:10} \t {1}'.format(name, desc))


def new_game():
    if len(sys.argv) != 3:
        print('Usage: pred new <game>')
        sys.exit()
    game = sys.argv[2]
    if os.path.exists(game):
        prompt = input('{} already exists. Overwrite? (Y or N): '.format(game))
        if prompt.upper() == 'Y':
            shutil.rmtree(game)
        else:
            sys.exit()
    os.mkdir(game)
    os.mkdir(game + '/backgrounds')
    os.mkdir(game + '/images')
    os.mkdir(game + '/sounds')
    os.mkdir(game + '/actors')
    file = open(game + '/game.py', 'w')
    file.write('WIDTH = 30\n')
    file.write('HEIGHT = 20\n')
    file.write('TITLE = \'Simple Game\'\n')
    file.close()



# shapes
RECT = 'rect' #:
SQUARE = 'rect'
CIRCLE = 'circle' #:
ELLIPSE = 'ellipse' #:

# colors
BLACK = (0, 0, 0) #:
SILVER = (192, 192, 192) #:
GRAY = (128, 128, 128) #:
WHITE = (255, 255, 255) #:
RED = (255, 0, 0) #:
MAROON = (128, 0, 0) #:
ORANGE = (255, 165, 0) #:
YELLOW = (255, 255, 0) #:
OLIVE = (128, 128, 0) #:
LIME = (0, 255, 0) #:
GREEN = (0, 128, 0) #:
AQUA = (0, 255, 255) #:
TEAL = (0, 128, 128) #:
BLUE = (0, 0, 255) #:
NAVY = (0, 0, 128) #:
CYAN = (0, 200, 200) #:
PINK = (255, 0, 255) #:
PURPLE = (128, 0, 128) #:

# Actor actions (omnidirectional)
IDLE = 'idle'
JUMP = 'jump'
RUN = 'run'
WALK = 'walk'
THROW = 'throw'
HURT = 'hurt'
DIE = 'die'
ATTACK = 'attack'
SHOOT = 'shoot'
HAPPY = 'happy'
DANCE = 'dance'

# Actor actions (directional)
ATTACK_BACK = 'attack_back'
ATTACK_FRONT = 'attack_front'
ATTACK_LEFT = 'attack_left'
ATTACK_RIGHT = 'attack_right'

SHOOT_BACK = 'shoot_back'
SHOOT_FRONT = 'shoot_front'
SHOOT_LEFT = 'shoot_left'
SHOOT_RIGHT = 'shoot_right'

THROW_BACK = 'throw_back'
THROW_FRONT = 'throw_front'
THROW_LEFT = 'throw_left'
THROW_RIGHT = 'throw_right'

DIE_BACK = 'die_back'
DIE_FRONT = 'die_front'
DIE_LEFT = 'die_left'
DIE_RIGHT = 'die_right'

FLY_BACK = 'fly_back'
FLY_FRONT = 'fly_front'
FLY_LEFT = 'fly_left'
FLY_RIGHT = 'fly_right'

JUMP_BACK = 'jump_back'
JUMP_FRONT = 'jump_front'
JUMP_LEFT = 'jump_left'
JUMP_RIGHT = 'jump_right'

IDLE_BACK = 'idle_back'
IDLE_FRONT = 'idle_front'
IDLE_LEFT = 'idle_left'
IDLE_RIGHT = 'idle_right'

IDLE_AIM_BACK = 'idle_aim_back'
IDLE_AIM_FRONT = 'idle_aim_front'
IDLE_AIM_LEFT = 'idle_aim_left'
IDLE_AIM_RIGHT = 'idle_aim_right'

WALK_BACK = 'walk_back'
WALK_FRONT = 'walk_front'
WALK_LEFT = 'walk_left'
WALK_RIGHT = 'walk_right'

WALK_AIM_BACK = 'walk_aim_back'
WALK_AIM_FRONT = 'walk_aim_front'
WALK_AIM_LEFT = 'walk_aim_left'
WALK_AIM_RIGHT = 'walk_aim_right'

RUN_BACK = 'run_back'
RUN_FRONT = 'run_front'
RUN_LEFT = 'run_left'
RUN_RIGHT = 'run_right'

DANCE_BACK = 'dance_back'
DANCE_FRONT = 'dance_front'
DANCE_LEFT = 'dance_left'
DANCE_RIGHT = 'dance_right'

LEFT = 'left'
RIGHT = 'right'
BACK = 'back'
FRONT = 'front'
POSE = 'pose'

OBSTACLE = 'wall'
GRAVITY = 0

# other contants
FOREVER = -1

# positions for scoring
UPPER_LEFT = 0
UPPER_RIGHT = 1
LOWER_LEFT = 2
LOWER_RIGHT = 3

ACCUMULATE = 0
TIMER = 1
VALUE = 2

WIN = 1
LOSE = 2



import pygame, random, math, sys
# implementation from http://www.learnpygame.com/2015/03/implementing-quadtree-collision-system.html

class Quadtree(object):
    def __init__(self, level, rect, sprites=[]):
        '''Quadtree box at with a current level, rect, list of sprites'''
        self.MAX_LEVEL = 4#max number of subdivisions
        self.level = level#current level of subdivision
        self.MAX_SPRITES = 3#max number of sprites without subdivision
        self.rect = rect#pygame rect object
        self.sprites = sprites#list of sprites
        self.branches = []#empty list that is filled with four branches if subdivided

    def get_rect(self):
        '''Returns quadtree rect object'''
        return self.rect

    def subdivide(self):
        '''Subdivides quadtree into four branches'''
        for rect in rect_quad_split(self.rect):
            branch = Quadtree(self.level+1, rect, [])
            self.branches.append(branch)

    def add(self, sprite):
        '''Adds a sprite to the list of sprites inside quadtree box'''
        self.sprites.append(sprite)

    def subdivide_sprites(self):
        '''Subdivides list of sprites in current box to four branch boxes'''
        for sprite in self.sprites:
            for branch in self.branches:
                if branch.get_rect().colliderect(sprite.rect):
                    branch.add(sprite)

    def test_collisions(self):
        '''Tests for collisions between all sprites'''
        for i, sprite in enumerate(self.sprites):
            for sprite2 in self.sprites[i+1:]:
                collide(sprite, sprite2)

    def update(self):
        '''Updates the quadtree and begins recursive process of subdividing or collision testing'''
        if len(self.sprites) > self.MAX_SPRITES and self.level <= self.MAX_LEVEL:
            self.subdivide()
            self.subdivide_sprites()
            for branch in self.branches:
                branch.update()
        else:
            self.test_collisions()



import sys, random, math, pygame
from functools import partial
from .utils import register_keydown, register_keyup, animate, randrange_float, sign, to_grid, is_wall, to_area, at, has_animation, get_animation, vsub, max_distance, has_obstacle
from .Globals import Globals
from .constants import *
import time
import copy


class Sprite():
    """

    The Predigame Sprite - a generic two-dimensional object that is integrated with other sprites in a larger scene. A sprite
    can consist of a bitmap (still) image or a basic geometrical shape (circle, rectangle, ellipse). Sprites in Predigame
    have some fun properties - they can be clicked, collide with other sprites, even fade, spin or pulse.

    """
    def __init__(self, surface, rect, tag=None, abortable=False, name=None):
        if len(Globals.instance.sprites) >= 9000:
            sys.exit('Too many sprites! You\'re trying to spawn over 9,000!')
        self.surface = surface.convert_alpha()
        self.origin_surface = self.surface
        self.rect = rect
        self.virt_rect = [float(self.rect.x), float(self.rect.y), float(self.rect.width), float(self.rect.height)]
        self.prev_rect = self.virt_rect
        self.surface = pygame.transform.scale(self.origin_surface, rect.size)
        self.mask = pygame.mask.from_surface(self.surface)
        self.needs_rotation = False
        self.move_speed = 5
        self.moving = False
        self.float_vec = (self.rect.x, self.rect.y)
        self.bounce_vec = (0, 0)
        self.sprite_scale_x = self.width
        self.sprite_scale_y = self.height
        self.rotate_angle = 0
        self.collisions = []
        self.clicks = []
        self.lifespan = -1
        self._pixelated = 0
        self._value = 0
        self.event_pos = None
        self.name = name
        self._tag = tag
        self.abortable = abortable
        self._mass = 0
        self.falling = False

        if tag not in Globals.instance.tags.keys():
            Globals.instance.tags[tag] = [self]
        else:
            Globals.instance.tags[tag].append(self)

    @property
    def pixelated(self):
        return self._pixelated

    @pixelated.setter
    def pixelated(self, value):
        self._pixelated = value
        self.pixelate(math.ceil(value))

    def mass(self, value):
        self._mass = value
        return self

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value
        return self



    @property
    def x(self):
        """
            get the x (left/right) position of the sprite

            :return: the x position of the sprite as an integer
        """
        return int(self.virt_rect[0] /Globals.instance.GRID_SIZE)

    @x.setter
    def x(self, value):
        """
            set the x (left/right) position of the sprite
        """
        self.needs_rotation = True
        self.virt_rect[0] = int(float(value) * Globals.instance.GRID_SIZE)

    @property
    def y(self):
        """
            get the y (up/down) position of the sprite

            :return: the y position of the sprite as an integer
        """
        return int(self.virt_rect[1] / Globals.instance.GRID_SIZE)

    @y.setter
    def y(self, value):
        """
            set the y (up/down) position of the sprite
        """
        self.needs_rotation = True
        self.virt_rect[1] = int(float(value) * Globals.instance.GRID_SIZE)

    @property
    def pos(self):
        """
            get the x and y positions as a tuple (pair of numbers).
            this is basically the same as calling `Sprite.x` and `Spite.y`
            separately.

            :return: a position tuple (x, y)
        """
        return self.x, self.y

    @pos.setter
    def pos(self, value):
        """
            set the x and y positions as a tuple (pair of numbers).
            this is basically the same as calling `Sprite.x = value` and `Sprite.y = value`
            separately.

            :return: a position tuple (x, y)
        """
        self.x = value[0]
        self.y = value[1]

    @property
    def width(self):
        """
            get the width of a sprite in terms of the grid block size.
        """
        return self.virt_rect[2] / Globals.instance.GRID_SIZE

    @property
    def height(self):
        """
            get the width of a sprite in terms of the grid block size.
        """
        return self.virt_rect[3] / Globals.instance.GRID_SIZE

    @property
    def size(self):
        """
            get the size of the sprite. sprites have a default size of 1 grid block.
            they can be increased or decreased as desired. the `Sprite.pulse()` effect
            will call size many times to create the desired effect.

            :todo: this method only returns the size along the x dimension. sprites should maintain separate sizes, one for each dimension.
        """
        return self.sprite_scale_x

    @size.setter
    def size(self, value):
        """
            set the size of the sprite. sprites have a default size of 1 grid block.
            they can be increased or decreased as desired. the `Sprite.pulse()` effect
            will call size many times to create the desired effect.

            :todo: this method applies the same magnification to both x and y dimensions. sprites should maintain separate sizes, one for each dimension.
        """
        self.needs_rotation = True
        self.sprite_scale_x = value
        self.sprite_scale_y = value

    def resize(self, value):
        """
            similar to `Sprite.size()` but callable as a method.
        """
        self.size = value
        return self

    @property
    def angle(self):
        """
            get the current rotated angle of the sprite

            :return: display angle in degrees
        """

        return self.rotate_angle

    @angle.setter
    def angle(self, value):
        """
            rotate the sprite by some `value` of degrees
        """
        self.needs_rotation = True
        self.rotate_angle = value

    @property
    def tag(self):
        """
            sprites can be "tagged" with extra information. for example,
            many zombie sprites can all have the singular tag "zombie".
            :return: the tag of the sprite (can be None if no tag provided)
        """
        return self._tag

    @tag.setter
    def tag(self, value):
        """
            sprites can be "tagged" with extra information. for example,
            many zombie sprites can all have the singular tag "zombie".
            (can be None if no tag provided)
        """
        self._tag = value


    @property
    def center(self):
        x, y, width, height = self.virt_rect[:]
        return (x + width/2.0)/ Globals.instance.GRID_SIZE, (y + height/2.0)/ Globals.instance.GRID_SIZE

    def stop(self):
        """ stop any running animations involving this sprite """
        for o in Globals.instance.animations:
            if o.obj == self:
                Globals.instance.animations.remove(o)
        return self

    def _update(self, delta):
        if self._mass > 0:
            self._update_gravity()
        if self.needs_rotation:
            self.rotate(0)
            self.needs_rotation = False
        self._handle_collisions()
        self.prev_rect = copy.copy(self.virt_rect)

    def _draw(self, surface):
        surface.blit(self.surface, self.rect)

    def _handle_collisions(self):
        for collision in self.collisions:
            if not collision['sprite'] in Globals.instance.sprites:
                self.collisions.remove(collision)
                continue

            if Globals.instance.PIXEL_COLLISIONS:
                offset = list(map(int,vsub(collision['sprite'].rect, self.rect)))
                overlap = self.mask.overlap_area(collision['sprite'].mask,offset)
                if overlap > 0:
                    collision['cb'](self, collision['sprite'])
            else:
                if self.rect.colliderect(collision['sprite'].rect):
                    collision['cb'](self, collision['sprite'])
                #break # only handle one collision per frame (for now)

    def _update_float(self, distance, time):
        float_x, float_y = self.float_vec
        time = self._calc_time((distance, distance))

        if self.virt_rect[0] == float_x:
            move_x = randrange_float(-distance, distance, distance * 2)
        else:
            move_x = -distance * sign(self.virt_rect[0] - float_x)

        if self.virt_rect[1] == float_y:
            move_y = randrange_float(-distance, distance, distance * 2)
        else:
            move_y = -distance * sign(self.virt_rect[1] - float_y)

        animate(self, time, partial(self._update_float, distance, time), x = self.x + move_x, y = self.y + move_y)

    def _update_bounce(self):
        if self.virt_rect[0] + self.virt_rect[2] >= Globals.instance.WIDTH and self.bounce_vec[0] > 0:
            self.bounce(True, False)
        elif self.virt_rect[0] <= 0 and self.bounce_vec[0] < 0:
            self.bounce(True, False)
        elif self.virt_rect[1] + self.virt_rect[3] >= Globals.instance.HEIGHT and self.bounce_vec[1] > 0:
            self.bounce(False, True)
        elif self.virt_rect[1] <= 0 and self.bounce_vec[1] < 0:
            self.bounce(False, True)

        x_dist = self.x
        if self.bounce_vec[0] >= 1:
            x_dist = (Globals.instance.WIDTH - (self.virt_rect[0] + self.virt_rect[2])) / (Globals.instance.GRID_SIZE * 1.0)
        y_dist = self.y
        if self.bounce_vec[1] >= 1:
            y_dist = (Globals.instance.HEIGHT - (self.virt_rect[1] + self.virt_rect[3])) / (Globals.instance.GRID_SIZE * 1.0)

        distance = min(x_dist, y_dist)
        time = self._calc_time((distance, distance))
        animate(self, time, self._update_bounce, x = self.x + distance * self.bounce_vec[0], y = self.y + distance * self.bounce_vec[1])

    def _handle_click(self, button, pos):
        for click in self.clicks:
            if button == click['btn']:
                self.event_pos = (1.0*pos[0])/Globals.instance.GRID_SIZE, (1.0*pos[1])/Globals.instance.GRID_SIZE
                click['cb'](self)

    def _calc_time(self, vector, position = None,speed = None):
        if not position:
            position = self.virt_rect[0:2]

        if not speed:
            speed = self.move_speed
        cur_x, cur_y = position
        new_x = cur_x + vector[0] * Globals.instance.GRID_SIZE
        new_y = cur_y + vector[1] * Globals.instance.GRID_SIZE
        distance = math.sqrt((new_x - cur_x)**2 + (new_y - cur_y)**2)
        time = (abs(distance) / speed) / 60.0
        return time

    def _complete_move(self, callback = None):
        self.moving = False
        if callback:
            callback()

    def destruct(self, time=1, **kwargs):
        """
            self destruct this sprite after `time` seconds

            :param time: the amount of time (in seconds) to stay alive before self destructing.

            :param callback: optional callback to execute after time is up (you'll need to do your own destructing)
        """

        self.lifespan = 0

        callback = kwargs.get('callback', None)
        if callback is None:
            callback = partial(self.destroy)
        animate(self, time, callback, lifespan=time)
        return self

    def fade(self, time=1, **kwargs):
        """
            make this sprite fade away and automatically self-destruct. this can be slow to run on some computers.

            :param time: the amount of time (in seconds) to run the fade effect.

            :todo: the final effect, like `destroy` should be configurable
        """
        #callback = kwargs.get('callback', None)
        animate(self, time, partial(self.destroy), pixelated = 100, size = 0)
        return self

    def move(self, vector, **kwargs):
        """
            move this sprite along a particular vector to the next grid cell (x and/or y dimensions)

            :param vector: the movement vector (x, y). `x` or `y` can take the values 0 (no movement), 1 (right or down movement), -1 (up or left movement). For example a vector of (1, 0) will move the sprite right and a vector of (1, -1) will move the sprite up and right.

            :param callback: an optional callback function to invoke when movement is complete.

            :param precondition: an optional callback function to invoke prior to a movement. preconditions all the sprite to avoid making a certain move (e.g. to avoid a wall)
        """
        self.moving = True
        callback = kwargs.get('callback', None)
        precondition = kwargs.get('precondition', None)
        action = kwargs.get('action', WALK)
        speed = kwargs.get('speed', None)
        x_dest = int(self.x + vector[0])
        y_dest = int(self.y + vector[1])
        time = self._calc_time(vector, speed=speed)

        x_dest, y_dest = max_distance(self.pos, (x_dest, y_dest), self.width, self.height)

        #if x_dest == self.x and y_dest == self.y:
        #    self._complete_move()
        #    return
        #print(' -- moving to ({},{})'.format(x_dest, y_dest))
        if precondition == None or precondition('move', self, vector):
            animate(self, time, partial(self._complete_move, callback), abortable=self.abortable, action=action, x = x_dest, y = y_dest)
        else:
            animate(self, time, partial(self._complete_move, callback), abortable=self.abortable, action=action, x = self.x, y = self.y)
        return self

    def move_to(self, *points, **kwargs):
        """
            move the sprite to a given point or set of points. movement will be set according to the `speed` of the sprite.

            :param points: one or more positions (x, y) pairs.

            :param callback: an optional callback function to invoke when movement is complete.


            Example:

            `s.move_to((0,0), (5,5), (10, 10), callback=s.destroy)`

            will move the sprite to position (0, 0) then (5,5) then (10,10) and then destroy itself.

        """
        #if self.moving:
        #    return self
        self.moving = True

        callback = partial(self._complete_move, kwargs.get('callback', None))
        action = kwargs.get('action', None)
        times = []

        for index, point in enumerate(points):
            if index == 0:
                prev_point = self.pos
            else:
                prev_point = points[index - 1]
            vector = (point[0] - prev_point[0], point[1] - prev_point[1])
            time = self._calc_time(vector, prev_point)
            times.append(time)

        for point in reversed(points):
            time = times.pop(-1)
            callback = partial(animate, self, time, callback, abortable=self.abortable, action=action, x = point[0], y = point[1])

        callback()
        return self

    def _update_gravity(self):
       # add bouncingness on the fall (should be based on the type of landing surface)

       direction_y = 1
       moving = False
       if self.prev_rect[1] > self.virt_rect[1]:
          direction_y = -1
          moving = True
       elif self.prev_rect[1] < self.virt_rect[1]:
          direction_y = 1
          moving = True

       if moving:
          return

       # animations for this sprite?
       ani = get_animation(self)

       cover_area = to_area(self.x, self.y, self.width, self.height, bottom_only=True)

       # see if any cover points are on the floor (not falling)
       #for p in cover_area:
       #   if p[1] >= Globals.instance.GRID_HEIGHT - 1:
       #      self.falling = False
       #      return

       # see if any the next points will hit an obstacle (coming down)
       next_area = list(map(lambda p : (p[0], p[1]+direction_y), cover_area))
       #print('cover {} ==> next {}'.format(cover_area, next_area))
       clear = True
       for p in next_area:
          if p in Globals.instance.cells:
             here = Globals.instance.cells[p]
             if has_obstacle(here):
                 clear = False
                 break

       #stationary and clear
       # - fall down to bottom
       if moving is False and clear and (ani is None or ani.action != JUMP):
           cb = None
           if ani is not None:
               cb = ani.callback
               Globals.instance.animations.remove(ani)
           self.move_to((self.x, Globals.instance.GRID_HEIGHT+5), action=GRAVITY, speed=30, callback=cb)
           return

       #falling and not clear
       # - stop
       #if moving and direction_y == 1 and clear is False:
       #      if ani is not None:
       #          print('aboring {} at ({},{})'.format(ani.attributes,self.x,self.y))
       #          Globals.instance.animations.remove(ani)
       ##          if ani.callback is not None:
       #              ani.callback()
       #      return

       #jumping and not clear
       # - stop
       #if moving and direction_y == -1 and clear is False:
       #      cb = None
       #      if ani is not None:
       #          cb = ani.callback
       #          Globals.instance.animations.remove(ani)
       #      self.move_to((self.x, (Globals.instance.GRID_HEIGHT-1)-self.height), action=GRAVITY, callback=cb)
       #      return

    def _continue_key(self, key, distance, **kwargs):
        p = kwargs.get('precondition', None)
        if key in Globals.instance.keys_pressed:
           self.move(distance, callback = partial(self._continue_key, key, distance, precondition = p), precondition = p)

    def _key_move(self, vector, **kwargs):
        # clean this up

        ani = get_animation(self)
        if ani is not None and (ani.action == JUMP or ani.action == GRAVITY):
           return

        if (self._mass == 0 or vector[1] == 0):
           self.move(vector, **kwargs)


    def keys(self, right = 'right', left = 'left', up = 'up', down = 'down', **kwargs):
        """
            make this sprite move to keyboard events. by default movements will be directional, by grid location -
            meaning that the sprite will move up, down, left, or right by one (default) grid cell.

            :param right: the keyboard button for right direction movement. default is the `right` arrow.

            :param left: the keyboard button for left direction movement. default is the `left` arrow.

            :param up: the keyboard button for up direction movement. default is the `up` arrow.

            :param down: the keyboard button for down direction movement. default is the `down` arrow.

            :param spaces: the number of grid cell. to move for each keyboard event. default is `1` grid cell.

            :param precondition: an optional callback function to invoke prior to a movement. preconditions all the sprite to avoid making a certain move (e.g. to avoid a wall)

            :param direction: the direction of the keyboard button and when to fire the event event. options are `down` for keyboard push down events or `up` for when the keyboard button is released. default is `down`.

        """
        p = kwargs.get('precondition', None)
        distance = kwargs.get('spaces', 1)

        register_key = None
        if kwargs.get('direction', 'down') == 'down':
            register_key = register_keydown
        else:
            register_key = register_keyup

        if right:
            register_key(right, partial(self._key_move, (1 * distance, 0), callback = partial(self._continue_key, right, (1 * distance, 0), precondition = p), precondition = p))
        if left:
            register_key(left, partial(self._key_move, (-1 * distance, 0), callback = partial(self._continue_key, left, (-1 * distance, 0), precondition = p), precondition = p))
        if up:
            register_key(up, partial(self._key_move, (0, -1 * distance), callback = partial(self._continue_key, up, (0, -1 * distance), precondition = p), precondition = p))
        if down:
            register_key(down, partial(self._key_move, (0, 1 * distance), callback = partial(self._continue_key, down, (0, 1 * distance), precondition = p), precondition = p))

        return self


    def follow(self):
        """
            tell the sprite to follow the mouse cursor as it moves around the screen
        """
        if self not in Globals.instance.mouse_motion:
            Globals.instance.mouse_motion.append(self)
        return self

    def speed(self, speed = 5):
        """
            set the speed of the sprite for movement actions. by default the sprite will move five grid cells per second.

            :param speed: the speed for movement actions. default is `5`.

        """
        self.move_speed = abs(speed)

        return self

    def float(self, distance = 0.25):
        """
            make the sprite float around on the screen. this provides a simple movement animation.

            :param distance: the amount of grid cell distance a sprite should float. default is `0.25`.

        """

        if self.moving:
            return self
        self.moving = True

        self.float_vec = (self.virt_rect[0], self.virt_rect[1])
        time = self._calc_time((distance, distance))

        animate(self, 0.016, partial(self._update_float, distance, time))

        return self

    def collides(self, sprites, callback):
        """
            register a callback function for when this sprite collides with another sprite. collision checks will occur as part of each update invocation.

            :param sprites: one (single object) or more (a list) of sprites to check for collisions.

            :param callback: the callback function to invoke when a collision is detected.

            :todo: confirm that collides are bi-directional events.

        """
        if not isinstance(sprites, list):
            sprites = [sprites]

        for sprite in sprites:
            if sprite == self:
                continue
            if sprite in self.collisions:
                self.collisions.remove(sprite)
            self.collisions.append({ 'sprite': sprite, 'cb': callback })

        return self

    def clicked(self, callback, button = 1):
        """
            register a callback function for when this sprite is clicked with a mouse button.

            :param callback: the callback function to invoke when a mouse click event is detected on this sprite.

            :param button: the mouse button to register for click events. options are left button `1` (default), right button `3` or middle button/track wheel `2`.

            :todo: confirm that collides are bi-directional events.

        """
        self.clicks.append({'btn': button, 'cb': callback})

        return self

    def pixelate(self, pixel_size = 10):
        """
            a simple effect for blurring or `pixelating` a sprite.

            :param pixel_size: the size of the pixels. default is `10`.
        """

        x_start = 0
        x_range = int(self.virt_rect[2])
        y_start = 0
        y_range = int(self.virt_rect[3])

        for x_offset in range(x_start, x_range, pixel_size):
            for y_offset in range(y_start, y_range, pixel_size):
                avg_color = pygame.transform.average_color(self.surface, (x_offset, y_offset, pixel_size, pixel_size))
                pygame.draw.rect(self.surface, avg_color, (x_offset, y_offset, pixel_size, pixel_size))

        return self

    def rotate(self, angle):
        """
            rotate the sprite by some angle (in degrees)

            :param angle: the directional angle (in degrees) to rotate this sprite.

        """
        self.rotate_angle += angle
        if self.rotate_angle > 360:
            self.rotate_angle -= 360

        x, y, width, height = self.virt_rect[:]
        center = (x + width/2), (y + height/2)

        scale_width = self.sprite_scale_x * Globals.instance.GRID_SIZE
        scale_height = self.sprite_scale_y * Globals.instance.GRID_SIZE
        self.surface = pygame.transform.scale(self.origin_surface, (int(scale_width), int(scale_height)))
        self.surface = pygame.transform.rotate(self.surface, self.rotate_angle)

        self.rect = self.surface.get_rect(center=center)
        new_rect = self.surface.get_rect()
        self.virt_rect[2] = new_rect[2]
        self.virt_rect[3] = new_rect[3]

        return self

    def flip(self, flip_x = True, flip_y = False):
        """
            flip the sprite image along x axis - left/right (default) or y axis - up/down.

            :param flip_x: flip the sprite along the x axis. default is `True`.

            :param flip_y: flip the sprite along the y axis. default is `False`.

        """

        self.origin_surface = pygame.transform.flip(self.origin_surface, flip_x, flip_y)
        self.surface = pygame.transform.flip(self.surface, flip_x, flip_y)

        return self

    def scale(self, size):
        """
            scale (increase or decrease) the size of the sprite. applies to both with and height.

            :param size: multiplier for modifying sprite scale. if greater than 1 sprite will increase in scale. if less than one (a decimal) sprite will decrease in case. if value is 1 sprite will left intact.

            :todo: scale needs to be applied separately to x and y axis.
        """

        if self.virt_rect[2] > Globals.instance.WIDTH and self.virt_rect[3] > Globals.instance.HEIGHT:
            return self
        width = self.virt_rect[2] * size
        height = self.virt_rect[3] * size

        self.virt_rect[2] = width
        self.virt_rect[3] = height
        self.sprite_scale_x = self.sprite_scale_x * size
        self.sprite_scale_y = self.sprite_scale_y * size

        self.surface = pygame.transform.smoothscale(self.origin_surface, (int(self.virt_rect[2]), int(self.virt_rect[3]))).convert_alpha()

        return self

    def bounce(self, bounce_x = True, bounce_y = True):
        x_dir, y_dir = self.bounce_vec
        if bounce_x:
            x_dir = -x_dir
        if bounce_y:
            y_dir = -y_dir
        self.bounce_vec = x_dir, y_dir

        return self

    def bouncy(self):
        """
            make this sprite a little bouncy. the `speed` attribute will control the speed of movement.
        """
        if self.moving:
            return self
        self.moving = True

        self.bounce_vec = (random.choice([-1, 1]), random.choice([-1, 1]))
        animate(self, 0.016, self._update_bounce)

        return self

    def pulse(self, time = 1, size = None):
        """
            add a pulsing effect. make the sprite simultaneously increase then decrease in size.

            :param time: the amount of time (in seconds) to complete a pulse. default is `1` second.

            :param size: the maximum size to pulse. default is to pulse until the size is doubled and then contract back to original size.

        """
        if not size:
            size = self.size * 2
        animate(self, time, partial(self.pulse, time, self.size), size = size)

        return self

    def spin(self, time = 1, rotations = None, **kwargs):
        """
            make this sprite spin.

            :param time: the amount of time (in seconds) to complete a revolution. default is `1` second.

            :param rotations: the amount of time (in seconds) to run the spin effect. default will run forever.

            :param callback: a function to call back when done spinning (rotations must be specified)

        """


        #if self.moving and not kwargs.get('spinning', False):
        #    return self
        callback = kwargs.get('callback', None)
        self.moving = True
        if rotations is not None:
            if rotations > 0:
                animate(self, time, partial(self.spin, time, rotations-1, spinning = True, callback=callback), angle = self.angle + 360)
            else:
                self.angle = 0
                if callback is not None:
                    callback()

        else:
            animate(self, time, partial(self.spin, time, spinning = True), angle = self.angle + 360)

        return self

    def destroy(self, *args):
        """
            destroy this sprite and remove from the game canvas.
        """
        if self in Globals.instance.sprites:
           Globals.instance.sprites.remove(self)
        if self in Globals.instance.tags[self._tag]:
           Globals.instance.tags[self._tag].remove(self)
        return self

    def wander(self, callback, time = 1):
        """
            turn your sprite into a bot by allowing the sprite to wander around. how the sprite "wanders" is based on the `callback` function.

            :param callback: the callback function to determine how the sprite should wander. this function should call a move operation.

            :param time: the frequency (in seconds) in order to call the wander function. default is `1` second.

        """
        callback()

        animate(self, time, partial(self.wander, callback, time))
        return self



# Code pulled from https://raw.githubusercontent.com/boppreh/maze/master/maze.py
# License: https://github.com/boppreh/maze/blob/master/LICENSE
# -*- coding: utf-8 -*-
import random

# Easy to read representation for each cardinal direction.
N, S, W, E = ('n', 's', 'w', 'e')

class Cell(object):
    """
    Class for each individual cell. Knows only its position and which walls are
    still standing.
    """
    def __init__(self, x, y, walls):
        self.x = x
        self.y = y
        self.walls = set(walls)

    def __repr__(self):
        # <15, 25 (es  )>
        return '<{}, {} ({:4})>'.format(self.x, self.y, ''.join(sorted(self.walls)))

    def __contains__(self, item):
        # N in cell
        return item in self.walls

    def is_full(self):
        """
        Returns True if all walls are still standing.
        """
        return len(self.walls) == 4

    def _wall_to(self, other):
        """
        Returns the direction to the given cell from the current one.
        Must be one cell away only.
        """
        assert abs(self.x - other.x) + abs(self.y - other.y) == 1, '{}, {}'.format(self, other)
        if other.y < self.y:
            return N
        elif other.y > self.y:
            return S
        elif other.x < self.x:
            return W
        elif other.x > self.x:
            return E
        else:
            assert False

    def connect(self, other):
        """
        Removes the wall between two adjacent cells.
        """
        other.walls.remove(other._wall_to(self))
        self.walls.remove(self._wall_to(other))

class Maze(object):
    """
    Maze class containing full board and maze generation algorithms.
    """

    # Unicode character for a wall with other walls in the given directions.
    UNICODE_BY_CONNECTIONS = {'ensw': '┼',
                              'ens': '├',
                              'enw': '┴',
                              'esw': '┬',
                              'es': '┌',
                              'en': '└',
                              'ew': '─',
                              'e': '╶',
                              'nsw': '┤',
                              'ns': '│',
                              'nw': '┘',
                              'sw': '┐',
                              's': '╷',
                              'n': '╵',
                              'w': '╴'}

    def __init__(self, width=20, height=10):
        """
        Creates a new maze with the given sizes, with all walls standing.
        """
        self.matrix = None
        self.width = width
        self.height = height
        self.cells = []
        for y in range(self.height):
            for x in range(self.width):
                self.cells.append(Cell(x, y, [N, S, E, W]))

    def __getitem__(self, index):
        """
        Returns the cell at index = (x, y).
        """
        x, y = index
        if 0 <= x < self.width and 0 <= y < self.height:
            return self.cells[x + y * self.width]
        else:
            return None

    def neighbors(self, cell):
        """
        Returns the list of neighboring cells, not counting diagonals. Cells on
        borders or corners may have less than 4 neighbors.
        """
        x = cell.x
        y = cell.y
        for new_x, new_y in [(x, y - 1), (x, y + 1), (x - 1, y), (x + 1, y)]:
            neighbor = self[new_x, new_y]
            if neighbor is not None:
                yield neighbor

    def at(self, x, y):
        """
        Returns True if there is a wall at (x, y). Values outside the valid
        range always return false.

        """
        if 0 <= x < len(self.matrix[0]) and 0 <= y < len(self.matrix):
            return self.matrix[y][x] != ' '
        else:
            return False

    def _to_str_matrix(self):
        """
        Returns a matrix with a pretty printed visual representation of this
        maze. Example 5x5:

        OOOOOOOOOOO
        O       O O
        OOO OOO O O
        O O   O   O
        O OOO OOO O
        O   O O   O
        OOO O O OOO
        O   O O O O
        O OOO O O O
        O     O   O
        OOOOOOOOOOO
        """
        str_matrix = [['O'] * (self.width * 2 + 1)
                      for i in range(self.height * 2 + 1)]

        for cell in self.cells:
            x = cell.x * 2 + 1
            y = cell.y * 2 + 1
            str_matrix[y][x] = ' '
            if N not in cell and y > 0:
                str_matrix[y - 1][x + 0] = ' '
            if S not in cell and y + 1 < self.width:
                str_matrix[y + 1][x + 0] = ' '
            if W not in cell and x > 0:
                str_matrix[y][x - 1] = ' '
            if E not in cell and x + 1 < self.width:
                str_matrix[y][x + 1] = ' '

        return str_matrix

    def __repr__(self):
        """
        Returns an Unicode representation of the maze. Size is doubled
        horizontally to avoid a stretched look. Example 5x5:

        ┌───┬───────┬───────┐
        │   │       │       │
        │   │   ╷   ╵   ╷   │
        │   │   │       │   │
        │   │   └───┬───┘   │
        │   │       │       │
        │   └───────┤   ┌───┤
        │           │   │   │
        │   ╷   ╶───┘   ╵   │
        │   │               │
        └───┴───────────────┘
        """
        # Starts with regular representation. Looks stretched because chars are
        # twice as high as they are wide (look at docs example in
        # `Maze._to_str_matrix`).
        skinny_matrix = self._to_str_matrix()

        # Simply duplicate each character in each line.
        double_wide_matrix = []
        for line in skinny_matrix:
            double_wide_matrix.append([])
            for char in line:
                double_wide_matrix[-1].append(char)
                double_wide_matrix[-1].append(char)

        # The last two chars of each line are walls, and we will need only one.
        # So we remove the last char of each line.
        matrix = [line[:-1] for line in double_wide_matrix]

        def g(x, y):
            """
            Returns True if there is a wall at (x, y). Values outside the valid
            range always return false.

            This is a temporary helper function.
            """
            if 0 <= x < len(matrix[0]) and 0 <= y < len(matrix):
                return matrix[y][x] != ' '
            else:
                return False

        # Fix double wide walls, finally giving the impression of a symmetric
        # maze.
        for y, line in enumerate(matrix):
            for x, char in enumerate(line):
                if not g(x, y) and g(x - 1, y):
                    matrix[y][x - 1] = ' '

        # Right now the maze has the correct aspect ratio, but is still using
        # 'O' to represent walls.

        # Finally we replace the walls with Unicode characters depending on
        # their context.
        for y, line in enumerate(matrix):
            for x, char in enumerate(line):
                if not g(x, y):
                    continue

                connections = set((N, S, E, W))
                if not g(x, y + 1): connections.remove(S)
                if not g(x, y - 1): connections.remove(N)
                if not g(x + 1, y): connections.remove(E)
                if not g(x - 1, y): connections.remove(W)

                str_connections = ''.join(sorted(connections))
                # Note we are changing the matrix we are reading. We need to be
                # careful as to not break the `g` function implementation.
                matrix[y][x] = Maze.UNICODE_BY_CONNECTIONS[str_connections]

        self.matrix = matrix
        # Simple double join to transform list of lists into string.
        return '\n'.join(''.join(line) for line in matrix) + '\n'

    def randomize(self):
        """
        Knocks down random walls to build a random perfect maze.

        Algorithm from http://mazeworks.com/mazegen/mazetut/index.htm
        """
        cell_stack = []
        cell = random.choice(self.cells)
        n_visited_cells = 1

        while n_visited_cells < len(self.cells):
            neighbors = [c for c in self.neighbors(cell) if c.is_full()]
            if len(neighbors):
                neighbor = random.choice(neighbors)
                cell.connect(neighbor)
                cell_stack.append(cell)
                cell = neighbor
                n_visited_cells += 1
            else:
                cell = cell_stack.pop()

    @staticmethod
    def generate(width=20, height=10):
        """
        Returns a new random perfect maze with the given sizes.
        """
        m = Maze(width, height)
        m.randomize()
        return m


class MazeGame(object):
    """
    Class for interactively playing random maze games.
    """
    def __init__(self, maze):
        self.maze = maze or Maze.generate()

    def _get_random_position(self):
        """
        Returns a random position on the maze.
        """
        return (random.randrange(0, self.maze.width),
                random.randrange(0, self.maze.height))

    def _display(self, pos, value):
        """
        Displays a value on the screen from an x and y maze positions.
        """
        x, y = pos
        # Double x position because displayed maze is double-wide.
        console.set_display(y * 2 + 1, x * 4 + 2, value)

    def play(self):
        """
        Starts an interactive game on this maze, with random starting and goal
        positions. Returns True if the user won, or False if she quit the game
        by pressing "q".
        """
        player = self._get_random_position()
        target = self._get_random_position()

        while player != target:
            console.display(str(self.maze))
            self._display(player, '@')
            self._display(target, '$')

            key = console.get_valid_key(['up', 'down', 'left', 'right', 'q'])

            if key == 'q':
                return False

            direction, difx, dify = {'up': (N, 0, -1),
                                     'down': (S, 0, 1),
                                     'left': (W, -1, 0),
                                     'right': (E, 1, 0)}[key]

            current_cell = self.maze[player]
            if direction not in current_cell:
                player = (player[0] + difx, player[1] + dify)

        console.display('You win!')
        console.get_key()
        return True



from setuptools import setup, find_packages


setup(
    name = 'predigame',
    version = '1.2.12',
    description = 'An instructional game development platform',
    url = 'http://predigame.io',
    author = 'Predicate Academy',
    author_email = 'support@predigame.io',
    packages = find_packages(),
    install_requires = ['pygame', 'pillow', 'astar==0.9', 'PyGithub==1.38','bresenham==0.2'],
    include_package_data = True,
    entry_points = {
        'console_scripts': [
            'pred = predigame:bootstrap'
        ]
    }
)
